@@ -1,8 +1,9 @@
 #!/bin/python
 # -*- coding: utf-8 -*-
-"""Functions for model parsing yaml -> working model instance. Involves a lot of dynamic function definition...
+"""Functions for model parsing yaml into a working model instance. Involves a lot of dynamic function definition...
 """
 
+from copy import copy
 import yaml
 import re
 import os
@@ -15,15 +16,10 @@ from copy import deepcopy
 from grgrlib import load_as_module
 from inspect import getmembers, isfunction
 from jax.experimental.host_callback import id_print as jax_print
-from .utilities import grids, dists
+from .utilities import grids, dists, interp
 from .parser.compile_functions import *
 from .parser.checks import *
 
-jax.config.update("jax_enable_x64", True)
-# set number of cores for XLA
-os.environ["XLA_FLAGS"] = f"--xla_force_host_platform_device_count={os.cpu_count()}"
-
-
 # initialize model cache
 cached_mdicts = ()
 cached_models = ()
@@ -47,25 +43,25 @@ def parse(mfile):
     model = yaml.safe_load(mtxt)
     # create nice shortcuts
     model['path'] = mfile
-    model["pars"] = model["parameters"]
+    model["pars"] = model.get("parameters")
     model["vars"] = model["variables"]
 
     return model
 
 
-def eval_strs(vdict, pars=None, context=globals()):
+def eval_strs(vdict, context={}):
+    """Evaluate a dictionary of strings into a given context
+    """
 
     if vdict is None:
         return None
 
-    if pars:
-        exec(
-            f"{', '.join(pars.keys())} = {', '.join(str(p) for p in pars.values())}", context)
-
     for v in vdict:
-        exec(f'{v} = {vdict[v]}', context)
         if isinstance(vdict[v], str):
-            vdict[v] = eval(f'{v}')
+            context[v] = eval(vdict[v], context)
+            vdict[v] = eval(v, context)
+        else:
+            context[v] = vdict[v]
 
     return vdict
 
@@ -92,18 +88,30 @@ def parse_external_functions_file(model):
     return
 
 
+def initialize_context():
+    context = globals().copy()
+    default_funcs = {'log': jnp.log,
+                     'exp': jnp.exp,
+                     'sqrt': jnp.sqrt,
+                     'max': jnp.maximum,
+                     'min': jnp.minimum}
+
+    context.update(default_funcs)
+    return context
+
+
 def load_external_functions_file(model, context):
     """Load the functions file as a module.
     """
 
     try:
         # load as a module
-        context['module'] = load_as_module(model["functions_file"])
+        module = load_as_module(model["functions_file"])
 
         def func_or_compiled(func): return isinstance(
             func, jaxlib.xla_extension.CompiledFunction) or isfunction(func)
         for m in getmembers(module, func_or_compiled):
-            exec(f'{m[0]} = module.{m[0]}', context)
+            context[m[0]] = m[1]
 
     except KeyError:
         pass
@@ -111,28 +119,37 @@ def load_external_functions_file(model, context):
     return False
 
 
-def compile_init_values(evars, decisions_inputs, initvals, stst):
+def compile_init_values(evars, pars, initvals, stst):
     """Combine all available information in initial guesses.
     """
 
+    ufixed_vars = {v: 1.1 for v in evars if v not in stst}
+    ufixed_pars = {v: 1.1 for v in pars if v not in stst}
+
     # get inital values to test the function
-    init = jnp.ones(len(evars)) * 1.1
+    init = {**ufixed_vars, **ufixed_pars}
 
     # structure: aggregate values first, then values of decisions functions
     if initvals is not None:
-        for v in initvals:
+        for v in init:
             # assign aggregate values
-            if v in evars:
-                init = init.at[evars.index(v)].set(initvals[v])
-
-    if stst:
-        for v in stst:
-            init = init.at[evars.index(v)].set(stst[v])
+            if v in initvals:
+                init[v] = initvals[v]
 
     return init
 
 
+def define_subdict_if_absent(parent, sub):
+    try:
+        return parent[sub]
+    except KeyError:
+        parent[sub] = {}
+        return parent[sub]
+
+
 def define_function(func_str, context):
+    """Define functions from string. Writes the function into a temporary file in order to get meaningful debug traces.
+    """
 
     # use a termporary file to get nice debug traces if things go wrong
     tmpf = tempfile.NamedTemporaryFile(mode="w", delete=False)
@@ -146,8 +163,10 @@ def define_function(func_str, context):
 
 
 def get_exog_grid_var_names(distributions):
-    # TODO: this will be important when implementing that grid parameters are endogenous variables
-    # TODO: when activated, backward calls already return exogenous grid vars (exog_grid_var). They are not yet stacked, and not yet an input to forward calls
+    """WIP. So far unused.
+    """
+    # NOTE: this will be important when implementing that grid parameters are endogenous variables
+    # NOTE: when activated, backward calls already return exogenous grid vars (exog_grid_var). They are not yet stacked, and not yet an input to forward calls
 
     exog_grid_var_names = ()
 
@@ -166,11 +185,24 @@ def get_exog_grid_var_names(distributions):
 def load(
     model,
     raise_errors=True,
-    lti_max_iter=500,
-    verbose=True,
-    testing=False,
+    verbose=True
 ):
-    """load model from dict or yaml file. Warning: contains filthy code (eg. globals, exec, ...)"""
+    """Load model from dict or yaml file.
+
+    Parameters
+    ----------
+    model : dict or string
+        either a dictionary or the path to a yaml file to be parsed
+    raise_errors : bool, optional
+        whether to raise errors while checking. False will let the model fail siliently for debugging. Defaults to True
+    verbose : bool, optional
+        inform that parsing is done. Defaults to True
+
+    Returns
+    -------
+    model : PizzaModel instance
+        The parsed model
+    """
 
     from .__init__ import PizzaModel
 
@@ -190,20 +222,16 @@ def load(
     # check if model is already cached
     if model in cached_mdicts:
         model = cached_models[cached_mdicts.index(model)]
-        model['context'] = globals()
-        load_external_functions_file(model, model['context'])
         print("(parse:) Loading cached model.")
         return model
 
-    model['context'] = globals()
+    model['context'] = initialize_context()
     load_external_functions_file(model, model['context'])
 
     defs = model.get("definitions")
-    # never ever use real numpy
-    if defs is not None:
-        for d in defs:
-            d = d.replace(" numpy ", " jax.numpy ")
-            exec(d, model['context'])
+    defs = '' if defs is None else defs
+    defs = '\n'.join(defs) if isinstance(defs, list) else defs
+    exec(defs, model['context'])
 
     eqns = model["equations"].copy()
 
@@ -216,9 +244,14 @@ def load(
     grids.create_grids(model.get('distributions'), model["context"])
 
     shocks = model.get("shocks") or ()
-    par = eval_strs(model["parameters"])
-    model["steady_state"]['fixed_evalued'] = stst = eval_strs(
-        model["steady_state"].get("fixed_values"), pars=par)
+    _ = define_subdict_if_absent(model, "steady_state")
+    fixed_values = define_subdict_if_absent(
+        model["steady_state"], "fixed_values")
+    stst = eval_strs(fixed_values, context=model['context'])
+    model["steady_state"]['fixed_evalued'] = stst
+
+    par = define_subdict_if_absent(model, "parameters")
+    model["parameters"] = par
     model["root_options"] = {}
 
     # collect number of foward and backward looking variables
@@ -227,18 +260,11 @@ def load(
     model["no_bwd"] = sum(var + "Lag" in "".join(model["equations"])
                           for var in evars)
 
-    stst_eqns = model["steady_state"].get("equations") or []
-    # add fixed values to the steady state equations
-    if stst is not None:
-        for key in stst:
-            # setting ALL occurences of the variable should be fine since we are using pinv later
-            stst_eqns.append(f"{key} = {stst[key]}")
-
     tmpf_names = ()
     # initialize storage for all function strings
     model['func_strings'] = {}
 
-    # TODO: currently disabled
+    # NOTE: currently disabled
     exog_grid_var_names = get_exog_grid_var_names(model.get('distributions'))
 
     # get function strings for decisions and distributions, if they exist
@@ -263,34 +289,39 @@ def load(
         # execute them
         tmpf_names += define_function(func_stst_dist_str, model['context']),
         tmpf_names += define_function(func_dist_str, model['context']),
-
     else:
         dist_names = []
 
     # collect initial guesses
-    model["init"] = compile_init_values(evars, decisions_inputs, eval_strs(
-        model["steady_state"].get("init_guesses")), stst)
+    init_guesses = eval_strs(model["steady_state"].get(
+        "init_guesses"), context=model['context'])
+    model["init"] = init = compile_init_values(evars, par, init_guesses, stst)
 
     # get strings that contains the function definitions
     model['func_strings']["func_pre_stst"] = compile_stst_func_str(
-        evars, eqns, par, stst_eqns)
+        evars, par, stst, init)
     model['func_strings']["func_eqns"] = compile_eqn_func_str(evars, deepcopy(eqns), par, eqns_aux=model.get(
         'aux_equations'), shocks=shocks, distributions=dist_names, decisions_outputs=decisions_outputs)
 
+    # test if model works. Writing to tempfiles helps to get nice debug traces if not
     tmpf_names += define_function(model['func_strings']
                                   ["func_eqns"], model['context']),
     tmpf_names += define_function(model['func_strings']
                                   ['func_pre_stst'], model['context']),
 
-    # test if model works. Writing to tempfiles helps to get nice debug traces if not
+    # get the initial decision functions
     if model.get('decisions'):
-        # try if function works on initvals
-        model['init_vf'] = model['steady_state']['init_guesses'][model['decisions']
-                                                                 ['inputs'][0]]  # let us for now assume that this must be present
+        init_vf_list = [model['steady_state']['init_guesses'][dec_input]
+                        for dec_input in model['decisions']['inputs']]  # let us for now assume that this must be present
+        model['init_vf'] = jnp.array(init_vf_list)
 
+        # check if initial decision functions and the distribution have same shapes
+        check_shapes(model['distributions'], model['init_vf'], dist_names)
+
+    # try if function works on initvals
     model['init_run'] = {}
     try:
-        check_func(model, shocks, par)
+        check_initial_values(model, shocks, par)
     except:
         if raise_errors:
             raise
